/*----------------------------------------------------------------------*/
/*                                                                      */
/* inflog.c : Written 1998 by Jochen Bauer                              */
/*                                  <jtb@theo2.physik.uni-stuttgart.de> */
/*                                                                      */
/* Disclaimer:                                                          */
/* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND WITHOUT ANY    */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE    */ 
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR   */
/* PURPOSE.                                                             */
/*                                                                      */
/* You may freely spread this file and any modification of it under one */
/* condition: If you mention my name then also mention the disclaimer.  */
/*                                                                      */
/* This file uses spoofit_v3.h written by Brecht Claerhout              */
/*                                           <Coder@reptile.rug.ac.be>  */
/*                                                                      */
/*----------------------------------------------------------------------*/ 
/*                                                                      */
/* 1. What is inflog?                                                   */    
/* inflog (=invalid flag logging) is a network tool for detecting QueSO */
/* scans. *It will only compile under Linux.*                           */
/*                                                                      */
/* Background:                                                          */
/* QueSO is a tool that uses TCP packets with invalid flag combinations */
/* to determine the operating system of the target host. As the three-  */
/* way handshake with the target host is never completed logging tools  */
/* like tcpwrapper will not be able to log this scan.                   */
/*                                                                      */
/* 2. How do I compile inflog?                                          */
/* Edit the LOGFILE and IFACE definitions if needed, then put this file */
/* together with spoofit_v3.h in the same directory an simply type:     */
/*                                                                      */
/*         gcc -O3 -o inflog inflog.c                                   */
/*                                                                      */
/* This program will only compile under Linux!                          */
/*                                                                      */
/* 3. How do I use inflog?                                              */
/* Become root and execute inflog. It will detach from the terminal,    */
/* open the logfile and log all TCP packets with invalid flags that are */   
/* used by QueSO. Because the interface is put into promiscuous mode,   */
/* the machine inflog runs on need not be the target of the QueSO scan  */
/* to log it, it only needs to "see" the packets QueSO sends. A typical */
/* QueSO scan will generate logfile entries like the following ones:    */     
/*                                                                      */
/* Sun Sep 20 18:21:33 Logging started....                              */
/* Sun Sep 20 18:27:56 x.x.x.x:8641-->y.y.y.y:23 FLAGS:F                */
/* Sun Sep 20 18:27:56 x.x.x.x:8643-->y.y.y.y:23 FLAGS:SF               */
/* Sun Sep 20 18:27:56 x.x.x.x:8644-->y.y.y.y:23 FLAGS:P                */
/* Sun Sep 20 18:27:56 x.x.x.x:8645-->y.y.y.y:23 FLAGS:YXS              */
/* Sun Sep 20 18:27:56 QueSO scan from x.x.x.x:8645 to y.y.y.y:23 detec */
/* ted                                                                  */
/*                                                                      */
/* I suggest that you get QueSO from www.rootshell.com or               */
/* www.apostols.org/projectz/queso/ and read it's documentation for     */
/* further background information                                       */
/*                                                                      */
/* 4. Performance issues                                                */
/* inflog will read each IP packet available on the interface and then  */
/* check the protocol number. If it's not a TCP packet it is discarded  */
/* otherwise the flags are analysed. On a heavily loaded network this   */
/* will mean a lot of work for inflog... To estimate the needed system  */
/* resources I did two tests:                                           */
/* I ping-flooded the host where inflog was running on (network is a    */
/* 100Mbit Ethernet). On a AMD K6 CPU (400.59 bogomips) inflog was now  */
/* consuming about 10% of the CPU time.                                 */
/* Next, I syn-flooded port 80 on the host where inflog was running on. */
/* The average CPU time consumption of inflog was now about 20%. When I */
/* tried to scan this host on port 80 while syn-flooding, QueSO was in  */
/* most cases not able to determine the operating system although port  */ 
/* 80 was open. However, the scans were entirely logged.                */
/* So, in summary, even an explicit DoS against the logging host could  */
/* not prevent inflog from doing it's job :-). It should be no problem  */
/* to give inflog a higher priority on loaded systems to ensure the     */ 
/* correct logging of scans even in cases of highest network load.      */
/*                                                                      */
/*----------------------------------------------------------------------*/
                       

#include "spoofit_v3.h"
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#define IFACE "eth0" /*the interface device*/ 
#define MASK 255  
#define LOGFILE "/var/log/inflog.log" 
#define FIN 1
#define SYN 2
#define PSH 8
#define XXX 64
#define YYY 128 /*some TCP flags, XXX and YYY are normally unused*/  

main()
{
int fd,f,alert,k,l;
char buffer[1500],dummy[32];
int tcpstart;
char proto,c;
struct IP_header *iphead;
struct TCP_header *tcphead;
unsigned short source_port;
unsigned short dest_port;
time_t now;
FILE *log;

DEV_PREFIX=14; 
alert=0; 
c=0;

log=fopen(LOGFILE,"a");        /*open logfile*/
fd=open_receiving(IFACE,0);    /*open receiving socket and put interface*/
			       /*into promiscuous mode*/
now=time(NULL);                      /*get time for logfile entry*/
      memcpy(dummy,ctime(&now),19);  /*cut the year off*/
      dummy[19]=0;
      fprintf(log,"%s ",dummy);
fprintf(log,"Logging started....\n");fflush(log);

l=fork(); /*fork and let the parent exit*/
  if (l>0)
   {
    close(fd);
    fclose(log);
    exit(0);
   }

signal(SIGHUP,SIG_IGN); /*ignore sighup*/
k=open("/dev/tty", O_RDWR|O_NOCTTY); /*detach from controlling tty*/
  if (k >= 0)
   {
    (void)ioctl(k,TIOCNOTTY,NULL);
    close(k);
   }
  else
   {
    perror("detach");
    exit(1);
   }

while(1)                           /*packet sniffing loop*/
 {
  get_packet(fd,buffer,&tcpstart,&proto);
  if(proto==6)    /*TCP packet?*/
   {
    tcphead=(struct TCP_header *)(buffer+tcpstart);     
    f=ntohs(tcphead->offset_flag)&MASK; /*get the flag bits*/ 
					/*for FIN,SYN,RST,PSH,ACK,XXX,YYY*/
    
    switch(f)
     {
     case FIN:alert=1;c=c|1;break;            /*FIN*/
     case SYN|FIN:alert=1;c=c|2;break;        /*SYN+FIN*/
     case PSH:alert=1;c=c|4;break;            /*PSH*/
     case SYN|XXX|YYY:alert=1;c=c|8;break;    /*SYN+unused flags*/
     default:alert=0;break;
     }
     
    if(alert==1)
     {
      iphead=(struct IP_header *)buffer;
      source_port=ntohs(tcphead->source);    /*packet source port*/
      dest_port=ntohs(tcphead->destination); /*packet dest. port*/
    
      now=time(NULL);                /*get time for logfile entry*/
      memcpy(dummy,ctime(&now),19);  /*cut the year off*/
      dummy[19]=0;
      fprintf(log,"%s ",dummy);
      fprintf(log,"%s:%i-->",inet_ntoa(iphead->source),source_port);
      fprintf(log,"%s:%i",inet_ntoa(iphead->destination),dest_port);
      
      fprintf(log," FLAGS:");
      if(f==FIN)
       fprintf(log,"F\n");
      if(f==(SYN|FIN))
       fprintf(log,"SF\n");     /*write the flags to the logfile*/
      if(f==PSH)
       fprintf(log,"P\n");
      if(f==(SYN|XXX|YYY))
       fprintf(log,"YXS\n");
      
      if(c==15)      /*scan completed?*/   
       {
	fprintf(log,"%s ",dummy); /*yes-->make logfile entry*/
        fprintf(log,"QueSO scan from %s:%i to %s:%i detected\n",
                       inet_ntoa(iphead->source),source_port,
                       inet_ntoa(iphead->destination),dest_port); 
        c=0;
       }        
      fflush(log);
      alert=0;
   
     }
   }
 }
}
